from unittest import TestCase

from holour import json_encode, json_decode
from holour.msg import RobotState, RobotStateChange


class TestRobotState(TestCase):

    def test_robot_state(self):
        robot_state = RobotState("Playing", True, True, 0.5, 1)
        robot_state_string = json_encode(robot_state)
        expected_string = '{"_type": "robot_state", "state": "Playing", "moving": true, "active": true, "target_speed": 0.5, "actual_speed": 1}'

        self.assertIs(type(robot_state_string), str, f"Got: {type(robot_state_string)}. Expected {str}")
        self.assertEqual(robot_state_string, expected_string, f"Expected {expected_string}, got: {robot_state_string}")

        robot_state_string_decoded = json_decode(robot_state_string)
        assert type(robot_state_string_decoded) == RobotState, f"Got: {type(robot_state_string_decoded)}. Expected {RobotState}"
        assert robot_state_string_decoded == robot_state, "The decoded object must be equal to the encoded"

    def test_robot_state_equals(self):
        ps1 = RobotState("Paused", False, False, 1, 1)
        ps2 = RobotState("Paused", False, False, 1, 1)
        ps3 = RobotState("Pausing", True, False, 1, 1)

        assert ps1 == ps2
        assert ps1 != ps3
        assert ps1 != "not status"

    def test_robot_state_repr(self):
        ps = RobotState("Stopping", True, False, 1, 1)
        expected, got = 'actual', f'{ps}'

        assert expected in got, f"Expected {expected} in got: {got}"


class TestRobotStateChange(TestCase):

    def test_robot_state_change(self):
        robot_state_change = RobotStateChange(RobotStateChange.START, 1)
        robot_state_string = json_encode(robot_state_change)
        expected_string = '{"_type": "robot_state_change", "state": "start", "speed": 1}'

        self.assertIs(type(robot_state_string), str, f"Got: {type(robot_state_string)}. Expected {str}")
        self.assertEqual(robot_state_string, expected_string, f"Expected {expected_string}, got: {robot_state_string}")

        string_decoded = json_decode(robot_state_string)
        assert type(string_decoded) == RobotStateChange, f"Got: {type(string_decoded)}. Expected {RobotStateChange}"
        assert string_decoded == robot_state_change, "The decoded object must be equal to the encoded"



from unittest import TestCase

from holour import json_encode, json_decode
from holour.msg import Task, Pose, Vector3, TaskMessage


class Test(TestCase):

    def test_task(self):

        task = Task('uuid', 'name', 'product_uuid', Pose(Vector3(0.0, 0.0, 0.0), Vector3(0.0, 0.0, 0.0)),
                    [Task.ASSIGNED_TO_HUMAN], Task.ASSIGNED_TO_HUMAN)
        task_string = json_encode(task)
        expected_string = '{"_type": "task", "uuid": "uuid", "name": "name", "product_uuid": "product_uuid", ' \
                          '"pose": {"_type": "pose", "name": "", "position": {"_type": "vector3", "x": 0.0, ' \
                          '"y": 0.0, "z": 0.0}, "rotation": {"_type": "vector3", "x": 0.0, "y": 0.0, "z": 0.0}' \
                          ', "axis_angle": {"_type": "axis_angle", "angle": 0.0, "x": 1, "y": 0, "z": 0}}, ' \
                          '"agents": ["human"], "assigned_to": "human", "conditions": [], ' \
                          '"conditions_fulfilled": true, "can_undo": false, "description": "", "status": "waiting"}'

        self.assertIs(type(task_string), str, f"Got: {type(task_string)}. Expected {str}")
        self.assertEqual(task_string, expected_string, f"Expected {expected_string}, got: {task_string}")

        task_decoded = json_decode(task_string)
        assert type(task_decoded) == Task, f"Got: {type(task_decoded)}. Expected {Task}"
        assert task_decoded == task, "The decoded object must be equal to the encoded"

    def test_task_equals(self):
        task1 = Task('uuid', 'name', 'product_uuid', Pose(Vector3(0.0, 0.0, 0.0), Vector3(0.0, 0.0, 0.0)),
                     [Task.ASSIGNED_TO_HUMAN], Task.ASSIGNED_TO_HUMAN)
        task2 = Task('uuid', 'name', 'product_uuid', Pose(Vector3(0.0, 0.0, 0.0), Vector3(0.0, 0.0, 0.0)),
                     [Task.ASSIGNED_TO_HUMAN], Task.ASSIGNED_TO_HUMAN)
        task3 = Task('other', 'name', 'product_uuid', Pose(Vector3(0.0, 0.0, 0.0), Vector3(0.0, 0.0, 0.0)),
                     [Task.ASSIGNED_TO_HUMAN], Task.ASSIGNED_TO_HUMAN)

        assert task1 == task2
        assert task1 != task3
        assert task1 != "not status"

    def test_task_repr(self):
        task = Task('uuid', 'name', 'product_uuid', Pose(Vector3(0.0, 0.0, 0.0), Vector3(0.0, 0.0, 0.0)),
                    [Task.ASSIGNED_TO_HUMAN], Task.ASSIGNED_TO_HUMAN, description="Hard task!")
        expected, got = 'Hard task!', f'{task}'

        assert expected in got, f"Expected {expected} in got: {got}"


class TestTaskMessage(TestCase):

    def test_task_message(self):
        task_message = TaskMessage('uuid', 'hej')
        task_message_string = json_encode(task_message)
        expected_string = '{"_type": "task_message", "task_uuid": "uuid", "message": "hej"}'

        assert type(task_message_string) == str
        assert task_message_string == expected_string, f"Expected {expected_string}, got: {task_message_string}"

        task_message_decoded = json_decode(task_message_string)
        assert type(task_message_decoded) == TaskMessage, f"Got: {type(task_message_decoded)}. Expected {TaskMessage}"
        assert task_message_decoded == task_message, "The decoded object must be equal to the encoded"

    def test_task_message_equals(self):
        at1 = TaskMessage('uuid', 'hej')
        at2 = TaskMessage('uuid', 'hej')
        at3 = TaskMessage('uuid2', '')

        assert at1 == at2
        assert at1 != at3
        assert at1 != "not status"

    def test_task_message_repr(self):
        add_task = TaskMessage('uuid', '')
        expected, got = 'uuid', f'{add_task}'

        assert expected in got, f"Expected {expected} in got: {got}"



from typing import List
from holour.msg import Pose


class Task:

    STATUS_MISSING_PRECONDITION = "missing_precondition"
    STATUS_WAITING = "waiting"
    STATUS_IN_PROGRESS = "in_progress"
    STATUS_ALMOST_DONE = "almost_done"
    STATUS_COMPLETED = "completed"
    STATUS_UNDO = "undo"
    STATUS_ERROR = "error"

    ASSIGNED_TO_HUMAN = "human"
    ASSIGNED_TO_ROBOT = "robot"

    def __init__(self, uuid: str, name: str, product_uuid: str, pose: Pose, agents: List[str],
                 assigned_to: str, conditions: List[str] = None, conditions_fulfilled: bool = True,
                 can_undo: bool = False, description: str = "", status: str = '', _type: str = ''):
        conditions = conditions if conditions else []
        status = status if status else self.STATUS_MISSING_PRECONDITION if len(conditions) > 0 else self.STATUS_WAITING
        assert type(uuid) == str
        assert type(name) == str
        assert type(product_uuid) == str
        assert type(pose) == Pose
        assert type(agents) == list
        assert type(assigned_to) == str
        assert type(conditions) == list, f"Expected list type. Type of conditions: {type(conditions)}"
        assert type(conditions_fulfilled) == bool
        assert type(can_undo) == bool
        assert type(description) == str
        assert type(status) == str

        self._type = 'task'
        self.uuid = uuid
        self.name = name
        self.product_uuid = product_uuid
        self.pose = pose
        self.agents = agents
        self.assigned_to = assigned_to
        self.conditions = conditions
        self.conditions_fulfilled = conditions_fulfilled
        self.can_undo = can_undo
        self.description = description
        self.status = status

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Task):
            return other.uuid == self.uuid \
                   and other.product_uuid == self.product_uuid \
                   and other.pose == self.pose \
                   and other.agents == self.agents \
                   and other.assigned_to == self.assigned_to \
                   and other.conditions == self.conditions \
                   and other.conditions_fulfilled == self.conditions_fulfilled \
                   and other.can_undo == self.can_undo \
                   and other.description == self.description \
                   and other.status == self.status
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self):
        return f"<uuid={self.uuid},name={self.name},product_uuid={self.product_uuid},pose={self.pose}," \
               f"agents={self.agents},assigned_to={self.assigned_to},conditions={self.conditions}," \
               f"status={self.status},conditions_fulfilled={self.conditions_fulfilled},can_undo={self.can_undo}," \
               f"description={self.description}>"


class TaskMessage:

    def __init__(self, task_uuid: str, message: str, _type: str = ''):
        self._type = 'task_message'
        self.task_uuid = task_uuid
        self.message = message

    def __eq__(self, other: object) -> bool:
        if isinstance(other, TaskMessage):
            return other.task_uuid == self.task_uuid and other.message == self.message
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self):
        return f"<task_uuid={self.task_uuid},message={self.message}>"



import logging
import sys
from typing import List


def setup_log(level=logging.DEBUG, handlers: List[logging.Handler] = None) -> None:
    if not handlers:
        handlers = []
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(get_format())
    logging.basicConfig(level=level, handlers=[handler] + handlers)


def get_format() -> logging.Formatter:
    return logging.Formatter('[%(asctime)s] [%(levelname)s] [%(name)s:%(funcName)s] - %(message)s')


def get_json_format() -> logging.Formatter:
    return logging.Formatter(
        "{'time':'%(asctime)s', 'name': '%(name)s', 'func': '%(funcName)s', 'level': '%(levelname)s', "
        "'message': '%(message)s'}"
    )


def json_log_handler(filename: str) -> logging.Handler:
    file_handler = logging.FileHandler(filename)
    file_handler.setFormatter(get_json_format())
    return file_handler




class RobotState:

    def __init__(self, state: str, moving: bool, active: bool, target_speed: float, actual_speed: float, _type: str = ''):
        self._type = 'robot_state'
        self.state = state
        self.moving = moving
        self.active = active
        self.target_speed = target_speed
        self.actual_speed = actual_speed

    def __eq__(self, other: object) -> bool:
        if isinstance(other, RobotState):
            return other.state == self.state and \
                   other.moving == self.moving and \
                   other.active == self.active and \
                   other.target_speed == self.target_speed and \
                   other.actual_speed == self.actual_speed
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self):
        return f"<robot_state<state={self.state},moving={self.moving},active={self.active}," \
               f"target_speed={self.target_speed},actual_speed={self.actual_speed}>"

    @staticmethod
    def translate_state_code(state: int) -> (str, bool, bool):
        # Return: state-string, is_moving, is_active (the last two are something I made up)
        if state == 0:
            return "Stopping", True, False
        elif state == 1:
            return "Stopped", False, False
        elif state == 2:
            return "Playing", True, True
        elif state == 3:
            return "Pausing", True, False
        elif state == 4:
            return "Paused", False, False
        elif state == 5:
            return "Stopped", False, False
        else:
            return "Unknown", False, False


class RobotStateChange:
    START = 'start'
    PAUSE = 'pause'
    STOP = 'stop'

    def __init__(self, state: str, speed: float, _type: str = ''):
        assert state in [RobotStateChange.START, RobotStateChange.PAUSE, RobotStateChange.STOP]
        self._type = 'robot_state_change'
        self.state = state
        self.speed = speed

    def __eq__(self, other: object) -> bool:
        if isinstance(other, RobotStateChange):
            return other.state == self.state and \
                   other.speed == self.speed
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self):
        return f"<robot_state_change<state={self.state},speed={self.speed}>"



import math
from typing import List


class Vector3:

    epsilon = 0.03

    def __init__(self, x: float, y: float, z: float, _type: str = ''):
        self._type = 'vector3'
        self.x = x
        self.y = y
        self.z = z

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Vector3):
            return math.fabs(other.x - self.x) < self.epsilon and \
                   math.fabs(other.y - self.y) < self.epsilon and \
                   math.fabs(other.z - self.z) < self.epsilon
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __sub__(self, other):
        return Vector3(self.x - other.x, self.y - other.y, self.z - other.z)

    def __repr__(self):
        return f"Vector3<x={self.x}, y={self.y}, z={self.z}>"


class Pose:

    def __init__(self, position: Vector3, rotation: Vector3, name: str = '', axis_angle=None,  _type: str = ''):
        self._type = 'pose'
        self.name = name
        self.position = position
        self.rotation = rotation
        self.axis_angle = AxisAngle.from_rotation_vec(rotation)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Pose):
            return other.name == self.name and other.position == self.position and other.rotation == self.rotation
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self):
        return f"Pose<name={self.name}, position={self.position}, orientation={self.rotation}>"

    def to_dict(self) -> dict:
        return {'x': self.position.x, 'y': self.position.y, 'z': self.position.z,
                'rx': self.rotation.x, 'ry': self.rotation.y, 'rz': self.rotation.z}

    @staticmethod
    def from_dict(pose: dict) -> 'Pose':
        return Pose(Vector3(pose.get('x', 0.0), pose.get('y', 0.0), pose.get('z', 0.0)),
                    Vector3(pose.get('rx', 0.0), pose.get('ry', 0.0), pose.get('rz', 0.0)), name=pose.get('name', ''))

    @staticmethod
    def from_list(pose: List, name: str = "") -> 'Pose':
        return Pose(Vector3(pose[0], pose[1], pose[2]), Vector3(pose[3], pose[4], pose[5]), name=name)

    @staticmethod
    def pose_trans(p1: 'Pose', p2: 'Pose', new_name: str = '') -> 'Pose':
        # from: https://forum.universal-robots.com/t/pose-trans-and-pose-inv-functions-in-urcap/1257 (scroll a bit down)
        res = Pose.matrix_multiply(p1.get_transform_matrix(), p2.get_transform_matrix())

        angle = AxisAngle.from_rotation_matrix([
            [res[0][0], res[0][1], res[0][2]],
            [res[1][0], res[1][1], res[1][2]],
            [res[2][0], res[2][1], res[2][2]]
        ])
        if new_name:
            name = new_name
        elif p1.name:
            name = p1.name
        else:
            name = p2.name

        return Pose(
            Vector3(res[0][3], res[1][3], res[2][3]),
            angle.get_rotation_vector(),
            name
        )

    def get_transform_matrix(self):
        angle = AxisAngle.from_rotation_vec(self.rotation)
        rm = angle.get_rotation_matrix()
        return [
            [rm[0][0], rm[0][1], rm[0][2], self.position.x],
            [rm[1][0], rm[1][1], rm[1][2], self.position.y],
            [rm[2][0], rm[2][1], rm[2][2], self.position.z],
            [0, 0, 0, 1]
        ]

    @staticmethod
    def matrix_multiply(a, b):
        res = [[0 for x in range(len(a))] for y in range(len(b[0]))]
        for i in range(0, len(a)):
            for j in range(0, len(b[0])):
                s = 0
                for k in range(0, len(a[i])):
                    s += a[i][k] * b[k][j]
                res[i][j] = s
        return res


class AxisAngle:

    def __init__(self, angle: float, x: float, y: float, z: float, _type: str = ''):
        self._type = "axis_angle"
        self.angle = angle
        self.x = x
        self.y = y
        self.z = z

    @staticmethod
    def from_rotation_vec(rotation: Vector3) -> 'AxisAngle':
        angle = math.sqrt(math.pow(rotation.x, 2) + math.pow(rotation.y, 2) + math.pow(rotation.z, 2))
        if angle == 0:
            y = z = 0
            x = 1
        else:
            x = rotation.x / angle
            y = rotation.y / angle
            z = rotation.z / angle
        return AxisAngle(angle, x, y, z)

    @staticmethod
    def from_rotation_matrix(m) -> 'AxisAngle':
        epsilon = 0.01
        epsilon2 = 0.1
        if math.fabs(m[0][1] - m[1][0]) < epsilon and math.fabs(m[0][2] - m[2][0]) < epsilon and math.fabs(m[1][2] - m[2][1]) < epsilon:
            if math.fabs(m[0][1] + m[1][0]) < epsilon2 and math.fabs(m[0][2] + m[2][0]) < epsilon2 and math.fabs(m[1][2] + m[2][1]) < epsilon2 and math.fabs(m[0][0] + m[1][1] + m[2][2] - 3) < epsilon2:
                return AxisAngle(0, 1, 0, 0)

            angle = math.pi
            xx = (m[0][0] + 1) / 2
            yy = (m[1][1] + 1) / 2
            zz = (m[2][2] + 1) / 2
            xy = (m[0][1] + m[1][0]) / 4
            xz = (m[0][2] + m[2][0]) / 4
            yz = (m[1][2] + m[2][1]) / 4

            if xx > yy and xx > zz:
                if xx < epsilon:
                    x = 0
                    y = 0.7071
                    z = 0.7071
                else:
                    x = math.sqrt(xx)
                    y = xy / x
                    z = xz / x
            elif yy > zz:
                if yy < epsilon:
                    x = 0.7071
                    y = 0
                    z = 0.7071
                else:
                    y = math.sqrt(yy)
                    x = xy / y
                    z = yz / y
            else:
                if zz < epsilon:
                    x = 0.7071
                    y = 0.7071
                    z = 0
                else:
                    z = math.sqrt(zz)
                    x = xz / z
                    y = yz / z
            return AxisAngle(angle, x, y, z)
        s = math.sqrt(
            (m[2][1] - m[1][2]) * (m[2][1] - m[1][2]) +
            (m[0][2] - m[2][0]) * (m[0][2] - m[2][0]) +
            (m[1][0] - m[0][1]) * (m[1][0] - m[0][1])
        )
        if math.fabs(s) < 0.001:
            s = 1

        angle = math.acos((m[0][0] + m[1][1] + m[2][2] - 1) / 2)
        x = (m[2][1] - m[1][2]) / s
        y = (m[0][2] - m[2][0]) / s
        z = (m[1][0] - m[0][1]) / s
        return AxisAngle(angle, x, y, z)

    def get_rotation_vector(self) -> Vector3:
        return Vector3(self.x * self.angle, self.y * self.angle, self.z * self.angle)

    def get_rotation_matrix(self) -> List[List[float]]:
        if self.angle == 0:
            return [[1, 0, 0], [0, 1, 0], [0, 0, 1]]

        c = math.cos(self.angle)
        s = math.sin(self.angle)
        t = 1.0 - c

        m00 = c + self.x * self.x * t
        m11 = c + self.y * self.y * t
        m22 = c + self.z * self.z * t

        tmp1 = self.x * self.y * t
        tmp2 = self.z * s
        m10 = tmp1 + tmp2
        m01 = tmp1 - tmp2

        tmp1 = self.x * self.z * t
        tmp2 = self.y * s
        m20 = tmp1 + tmp2
        m02 = tmp1 - tmp2

        tmp1 = self.y * self.z * t
        tmp2 = self.x * s
        m21 = tmp1 + tmp2
        m12 = tmp1 - tmp2

        return [[m00, m01, m02], [m10, m11, m12], [m20, m21, m22]]


class Poses:

    def __init__(self, poses: List[Pose], connected: bool = False, _type: str = ''):
        self._type = 'poses'
        self.connected = connected
        self.poses = poses

    def add(self, pose: Pose):
        self.poses.append(pose)

    def convert_and_add_unity_pose(self, pose: Pose):
        self.poses.append(Pose(
            Vector3(pose.position.x, pose.position.z, pose.position.y),
            pose.rotation,  # TODO: add some conversion?
            pose.name
        ))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Poses):
            return other.connected == self.connected and other.poses == self.poses
        return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self):
        return f"Poses<connected={self.connected}, poses={self.poses}>"
